import collections

from django.db import models
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from filelock import FileLock

from redmin.utils import attr
from .base import RedminModel
from .domain import Domain
from .group import Group
from .user import User

cache = collections.defaultdict(dict)

lock = FileLock("redmin_permission.lock")
lock.release(force=True)


@receiver([post_delete, post_save])
def clear_permissions(sender, **kwargs):
    if sender in [Group, User] or issubclass(sender, Permission):
        with lock:
            cache.clear()


def _get_permissions():
    if not cache:
        with lock:
            domains = Domain.objects.all()
            for permission in GroupPermission.objects.all():
                for user in User.objects.filter(group=permission.group):
                    cache[user][permission.domain.get_model()] = permission
            for permission in UserPermission.objects.all():
                cache[permission.user][permission.domain.get_model()] = permission

            for permission in GroupPermission.objects.filter(domain__name=SuperPermissionModel.__name__):
                for user in User.objects.filter(group=permission.group):
                    for domain in domains:
                        model = domain.get_model()
                        if model not in cache[user]:
                            cache[user][model] = permission

            for permission in UserPermission.objects.filter(domain__name=SuperPermissionModel.__name__):
                for domain in domains:
                    model = domain.get_model()
                    if model not in cache[permission.user]:
                        cache[permission.user][model] = permission

            for user in User.objects.all():
                for domain in domains:
                    model = domain.get_model()
                    if model not in cache[user]:
                        cache[user][model] = UserPermission(user=user, domain=domain)

    return cache


class ModelPermission:
    def __init__(self, **kwargs):
        self.removable = kwargs.get("removable")
        self.cloneable = kwargs.get("cloneable")
        self.viewable = kwargs.get("viewable")
        self.savable = kwargs.get("savable")

    @property
    def operable(self):
        return self.viewable or self.removable or self.cloneable


class PermissionItem:
    creatable = "creatable"
    savable = "savable"
    removable = "removable"
    cloneable = "cloneable"
    exportable = "exportable"
    viewable = "viewable"
    listable = "listable"
    selectable = 'selectable'


VALID_ITEMS = set([key for key, _ in vars(PermissionItem).items() if not key.startswith("_")])


class Permission(RedminModel):
    class Meta:
        abstract = True

    domain = models.ForeignKey(Domain, on_delete=models.CASCADE)
    creatable = models.BooleanField("可创建", default=False)
    savable = models.BooleanField("可保存", default=False)
    removable = models.BooleanField("可删除", default=False)
    cloneable = models.BooleanField("可复制", default=False)
    exportable = models.BooleanField("可导出", default=False)
    viewable = models.BooleanField("可查看", default=False)
    listable = models.BooleanField("可列表", default=False)
    selectable = models.BooleanField("可选择", default=True)

    def to_instance_permission(self):
        return ModelPermission(
            removable=self.removable,
            cloneable=self.cloneable,
            viewable=self.viewable,
            savable=self.savable
        )

    @property
    def operable(self):
        return self.viewable or self.removable or self.cloneable

    @classmethod
    def get_permission(cls, user, model):
        if not user or not model:
            return None
        return _get_permissions()[user][model]

    @classmethod
    def has_permission(cls, user, model, item):
        if item in VALID_ITEMS:
            return attr(cls.get_permission(user, model), item)


class GroupPermission(Permission):
    class Meta:
        verbose_name_plural = verbose_name = "权限(用户组)"
        unique_together = (('group', 'domain'),)

    group = models.ForeignKey(Group, on_delete=models.CASCADE)


class UserPermission(Permission):
    class Meta:
        verbose_name_plural = verbose_name = "权限(用户)"
        unique_together = (('user', 'domain'),)

    user = models.ForeignKey(User, on_delete=models.CASCADE)


class SuperPermissionModel(RedminModel):
    """
    代表所有模型，一般用在admin用户组/用户下
    """

    class Meta:
        verbose_name_plural = verbose_name = "超級模型"



import datetime
import gzip

import xlwt
from django.http import HttpResponse
from django.utils.http import urlquote
from django.utils.six import BytesIO

from redmin.models import Permission, PermissionItem, Field
from redmin.utils import attr, display
from .list import AdminListView


class AdminExportView(AdminListView):
    paginate_by = None

    def has_permission(self):
        return Permission.has_permission(self.request.user, self.model, PermissionItem.exportable)

    def get_fields(self):
        return [field for field in Field.get_fields(self.request.user, self.model) if field.exportable]

    def get(self, request, *args, **kwargs):
        return self.export_excel(self.get_queryset())

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["view_type"] = 'export'
        return context

    def export_excel(self, objects):
        fields = self.get_fields()
        model_verbose = attr(self.model, '_meta.verbose_name')
        wb = xlwt.Workbook(encoding='utf-8')
        sheet = wb.add_sheet('%s列表' % model_verbose)
        style_heading = xlwt.easyxf("""
            font:name Arial,colour_index white,bold on,height 0xA0;
            align:wrap off,vert center,horiz center;
            pattern:pattern solid,fore-colour 0x19;
            borders:left THIN,right THIN,top THIN,bottom THIN;
            """)
        style_body = xlwt.easyxf("""
                font:name Arial,bold off,height 0XA0;
                align:wrap on,vert center,horiz left;
                borders:left THIN,right THIN,top THIN,bottom THIN;
                """)
        # style_green = xlwt.easyxf(" pattern: pattern solid,fore-colour 0x11;")
        # style_red = xlwt.easyxf(" pattern: pattern solid,fore-colour 0x0A;")
        fmts = [
            'M/D/YY', 'D-MMM-YY', 'D-MMM', 'MMM-YY',
            'h:mm AM/PM', 'h:mm:ss AM/PM', 'h:mm', 'h:mm:ss', 'M/D/YY h:mm', 'mm:ss', '[h]:mm:ss', 'mm:ss.0',
        ]
        style_body.num_format_str = fmts[0]

        for column, field in enumerate(fields):
            sheet.write(0, column, str(field.verbose), style_heading)

        for row, obj in enumerate(objects, 1):
            for column, field in enumerate(fields, 0):
                sheet.write(row, column, str(display(obj, field.name)), style_body)

        buf = BytesIO()
        wb.save(buf)
        buf.seek(0)
        response = HttpResponse(gzip.compress(buf.getvalue()), content_type="application/vnd.ms-excel")
        filename = model_verbose + "-" + datetime.datetime.now().strftime('%Y%m%d%H%M%S')
        response['Content-Disposition'] = 'attachment;filename=%s.xls' % urlquote(filename)
        response['Content-Encoding'] = 'gzip'
        return response



from .attrs import attr, display
from .common import to_json, first, last, is_windows, get_ip, get_domain, null_to_emtpy
from .decorators import auto_repr, auto_str, memorize, get_wrapper
from .imports import import_class, import_sub_classes, import_submodules, import_model
from .request import param, int_param, bool_param, json_response, get_url_prefix
from .string import is_empty, is_not_empty
from .template import render
from .validators import validate_mac_address



from .base import get_base_view_class, get_view,get_admin_domain_patterns,get_admin_login_path
from .create import AdminCreateView
from .delete import AdminDeleteView
from .export import AdminExportView
from .form import AdminFormView
from .list import AdminListView
from .update import AdminUpdateView




import json
from collections import OrderedDict

from django import forms
from django.contrib.messages.views import SuccessMessageMixin
from django.forms import ChoiceField

from redmin.models import Permission, PermissionItem, Field, Filter
from redmin.utils import attr
from .mixins import ContextMixin, AccessMixin


class AdminFormView(SuccessMessageMixin, ContextMixin, AccessMixin):

    def get_form_class(self):
        user = self.request.user
        attributes = [field.attribute for field in Field.get_fields(user, self.model) if "." not in field.attribute and field.formable]
        form = type(f"DynamicForm_Model_{self.model.__name__}_User_{user.id}", (forms.ModelForm,), dict(
            Meta=type("Meta", (), dict(
                model=self.model,
                fields=attributes,
            )),
            __module__=__name__
        ))
        return form

    def post(self, request, *args, **kwargs):
        if Permission.has_permission(self.request.user, self.model, PermissionItem.savable):
            return super().post(request, *args, **kwargs)
        else:
            return self.handle_no_permission()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["model"] = self.model
        context["model_name"] = self.model.__name__
        group_fields = Field.get_group_fields(self.request.user, self.model, has_domain=True, reverse=False)
        context["group_fields_json"] = json.dumps([[{'attribute': field.attribute, "class": field.attribute_domain.get_model().__name__} for field in fields] for fields in group_fields])
        self.add_foreign_field_choices(context["form"], attr(context, "object"), group_fields)
        return context

    def get_option_query(self, cls):
        return cls.objects

    def add_foreign_field_choices(self, form, obj, group_fields):
        choices = []
        for foreign_fields in group_fields:
            last_field: Field = None
            last_value = None
            foreign_fields = [field for field in foreign_fields if field.formable]
            for index in range(len(foreign_fields)):
                field = foreign_fields[index]
                cls = field.model
                queryset = None
                attribute = field.attribute
                if index == 0 or last_value or last_field.nullable:
                    queryset = self.get_option_query(cls)
                    if last_value:
                        x = last_field.attribute[len(attribute) + 1:]
                        if x:
                            filters = {f"{x}_id": last_value}
                            queryset = queryset.filter(**filters)

                queryset = Filter.filter(queryset, self.request, cls)
                options = [(e.id, str(e)) for e in queryset.all()] if queryset else []
                if len(options) > 1:
                    options = [('', '------')] + options
                value = attr(obj, f'{attribute}_id')
                choices.append((attribute, ChoiceField(
                    required=True if form.fields.pop(attribute, False) else False,
                    initial=value,
                    label=field.title,
                    choices=options
                )))

                last_value = value
                last_field = field

        fields = OrderedDict(choices)
        fields.update(form.fields)
        form.fields = fields



from django.db import models

from .base import RedminModel
from .domain import Domain


class Menu(RedminModel):
    class Meta:
        ordering = ["sequence"]
        verbose_name_plural = verbose_name = "菜单"

    name = models.CharField("名称", max_length=100, unique=True)
    sequence = models.IntegerField("排序", default=0)

    def __str__(self):
        return self.name


class DomainMenu(RedminModel):
    class Meta:
        ordering = ["menu__sequence", "domain__sequence"]
        verbose_name_plural = verbose_name = "模型菜单"
        unique_together = [('menu', 'domain')]

    menu = models.ForeignKey(Menu, on_delete=models.CASCADE)
    domain = models.ForeignKey(Domain, on_delete=models.CASCADE)

    def __str__(self):
        return f"DomainMenu({self.id}-{self.menu.name}-{self.domain.name})"



from rest_framework.authentication import BaseAuthentication, BasicAuthentication as RestBasicAuthentication


class SessionAuthentication(BaseAuthentication):

    def authenticate(self, request):
        user = getattr(request._request, 'user', None)
        if not user or not user.is_active:
            return None
        return (user, None)


class BasicAuthentication(RestBasicAuthentication):
    pass



from django.urls import include, path

from .pagination import RedminPagination
from .permission import RolePermission
from .view import account


def get_rest_patterns(prefix):
    from redmin.rest.router import BaseRouter
    from redmin.rest.serializer.base import BaseSerializer
    from redmin.rest.viewset.base import BaseViewSet, load_viewset_class
    from redmin.rest.utils import get_all_models
    r = BaseRouter()
    for model in get_all_models():
        viewset = load_viewset_class(model)
        r.register(model.__name__, viewset)

    account_methods = ["login", "logout", "token", "profile"]
    account_patterns = [path(f'{prefix}/account/{method}', getattr(account, method), name=f"account-{method}") for method in account_methods]
    return account_patterns + [
        path(f'{prefix}/auth/', include('rest_framework.urls')),
        path(f'{prefix}/', include(r.urls)),
    ]



from .base import RedminModel
from .choice import Choice
from .domain import Domain
from .field import Field, GroupField, UserField, CommonField
from .filter import UserFilter, GroupFilter, FilterValueType, Filter
from .group import Group
from .permission import GroupPermission, UserPermission, SuperPermissionModel, Permission, PermissionItem
from .user import User
from .menu import Menu, DomainMenu



from django.contrib.auth.models import AbstractUser
from django.db import models

from .base import RedminModel
from .group import Group


class User(AbstractUser, RedminModel):
    class Meta:
        verbose_name_plural = verbose_name = "用户"

    group = models.ForeignKey(Group, on_delete=models.CASCADE)
    title = models.CharField('名称', max_length=50, blank=False, null=False)

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        if not self.title:
            self.title = self.username
        return super().save(*args, **kwargs)



import logging

from rest_framework import serializers

from redmin.utils import import_class, attr

logger = logging.getLogger(__name__)


class BaseSerializer(serializers.ModelSerializer):
    def build_relational_field(self, field_name, relation_info):
        result = super().build_relational_field(field_name, relation_info)
        field_kwargs = result[1]
        # field_kwargs['queryset'] = filter_by_user(field_kwargs.pop("queryset"), attr(self, "context.request.user"))
        return result


def load_serializer_class(model):
    app = model.__module__.split(".")[0]
    name = f"{model.__name__}Serializer"
    try:
        clazz = import_class(f'{app}.rest.serializer.{name}')
        logger.info(f"load rest serializer:{clazz}")
        return clazz
    except (AttributeError, ModuleNotFoundError):
        model_fields = attr(model, '_meta.fields')
        all_fields = (set(attr(model, "form_fields", [])) | set(attr(model, "list_fields", []))) or set(
            [f.name for f in model_fields])
        excludes = set(attr(model, "form_exclude", [])) & set(attr(model, "list_exclude", []))
        fields = list(all_fields - excludes)
        return get_dynamic_serializer_class(model, fields)


def get_dynamic_serializer_class(model, fields):
    name = f"Dynamic{model.__name__}Serializer"
    return type(name, (BaseSerializer,), dict(
        Meta=type("Meta", (), dict(
            model=model,
            fields=fields,
            __module__=f"{__name__}.{name}"
        )),
        __module__=__name__
    ))



import base64

from Crypto import Random
from Crypto.Cipher import PKCS1_v1_5 as Cipher_pkcs1_v1_5
from Crypto.Hash import SHA
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5 as Signature_pkcs1_v1_5


def generate_public_private_key():
    rsa = RSA.generate(2048, Random.new().read)
    private_key = rsa.exportKey()
    public_key = rsa.publickey().exportKey()
    return private_key, public_key


def encrypt(message, public_key):
    rsakey = RSA.importKey(public_key)
    cipher = Cipher_pkcs1_v1_5.new(rsakey)
    cipher_text = base64.b64encode(cipher.encrypt(bytes(message.encode("utf8"))))
    return cipher_text


def decrypt(cipher_text, private_key):
    rsakey = RSA.importKey(private_key)
    cipher = Cipher_pkcs1_v1_5.new(rsakey)
    sentinel = {}
    text = cipher.decrypt(base64.b64decode(cipher_text), sentinel=sentinel)
    return str(text, "utf8")


def sign(message, private_key):
    rsakey = RSA.importKey(private_key)
    signer = Signature_pkcs1_v1_5.new(rsakey)
    digest = SHA.new()
    digest.update(message.encode("utf8"))
    signature = base64.b64encode(signer.sign(digest))
    return signature


def verify_sign(message, signature, public_key):
    rsakey = RSA.importKey(public_key)
    verifier = Signature_pkcs1_v1_5.new(rsakey)
    digest = SHA.new()
    digest.update(message.encode("utf8"))
    return verifier.verify(digest, base64.b64decode(signature))


def main():
    with open(r'D:\app\luo-video\key\priv.pem', 'rb+') as f:
        private_key = f.read()

    with open(r'D:\app\luo-video\key\pub.pem', 'rb+') as f:
        public_key = f.read()

    message = "hpPEbBjenqRrNTQwvErznOBf5DbPKubi"
    cipher = Cipher_pkcs1_v1_5.new(RSA.importKey(public_key))
    cipher_text = cipher.encrypt(bytes(message.encode("utf8")))
    result = cipher_text.hex()
    print("加密后:", result)

    cipher = Cipher_pkcs1_v1_5.new(RSA.importKey(private_key))
    print("解密数据:", cipher.decrypt(bytes.fromhex(result), {}))
    with open(r'/tmp/code.txt', 'w') as f:
        f.write(result)

    with open(r'/tmp/code.txt', 'r') as f:
        file_content = f.read()

    print("解密数据载入的数据:", cipher.decrypt(bytes.fromhex(file_content), {}))

    with open(r'D:\app\luo-video\tmp\3a6ca2557344d16951c3457fde2035c46704fff41b4e2d4dd21db8e133da0385.key', 'r') as f:
        code = f.read()

    print("载入数据:", code)
    print("解密载入的数据:", cipher.decrypt(bytes.fromhex(code), {}))


if __name__ == "__main__":
    main()



# Generated by Django 2.2.5 on 2019-09-20 17:08

from django.conf import settings
import django.contrib.auth.models
import django.contrib.auth.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('username', models.CharField(error_messages={'unique': 'A user with that username already exists.'}, help_text='Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.', max_length=150, unique=True, validators=[django.contrib.auth.validators.UnicodeUsernameValidator()], verbose_name='username')),
                ('first_name', models.CharField(blank=True, max_length=30, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=150, verbose_name='last name')),
                ('email', models.EmailField(blank=True, max_length=254, verbose_name='email address')),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('title', models.CharField(max_length=50, verbose_name='名称')),
            ],
            options={
                'verbose_name': '用户',
                'verbose_name_plural': '用户',
                'ordering': ['-id'],
            },
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='Domain',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=100, verbose_name='描述')),
                ('sequence', models.IntegerField(default=99999, verbose_name='排序')),
                ('app', models.CharField(max_length=100, verbose_name='应用')),
                ('name', models.CharField(max_length=100, verbose_name='名称')),
            ],
            options={
                'verbose_name': '模型',
                'verbose_name_plural': '模型',
                'ordering': ['sequence', 'id'],
                'unique_together': {('app', 'name')},
            },
        ),
        migrations.CreateModel(
            name='Field',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=200, verbose_name='显示名称')),
                ('sequence', models.IntegerField(verbose_name='序号')),
                ('listable', models.BooleanField(default=True, verbose_name='在列表中显示')),
                ('formable', models.BooleanField(default=True, verbose_name='在表单中显示')),
                ('sortable', models.BooleanField(default=True, verbose_name='可排序')),
                ('exportable', models.BooleanField(default=True, verbose_name='可导出')),
                ('nullable', models.BooleanField(default=False, verbose_name='可以为空')),
                ('unique', models.BooleanField(default=False, verbose_name='是否唯一性')),
                ('default', models.CharField(blank=True, max_length=500, null=True, verbose_name='默认值')),
                ('form_editable', models.BooleanField(default=True, verbose_name='表单可编辑')),
                ('list_editable', models.BooleanField(default=False, verbose_name='列表可编辑')),
                ('searchable', models.BooleanField(default=False, help_text='通过一个关键字搜索选中的每一列', verbose_name='可搜索')),
                ('filterable', models.BooleanField(default=True, help_text='是否可在列表任务栏过滤', verbose_name='可过滤')),
                ('serializable', models.BooleanField(default=True, help_text='RestAPI里是否可以序列化', verbose_name='可序列化')),
                ('help_text', models.TextField(blank=True, null=True, verbose_name='帮助文本')),
                ('attribute', models.CharField(max_length=100, verbose_name='字段名称')),
                ('group_sequence', models.IntegerField(verbose_name='分组序号')),
                ('python_type', models.CharField(max_length=200, verbose_name='原生类型')),
                ('data_type', models.CharField(max_length=10, verbose_name='数据类型')),
                ('max_length', models.IntegerField(default=0, verbose_name='最大长度')),
                ('attribute_domain', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='attribute_domain', to='redmin.Domain', verbose_name='字段模型')),
                ('base', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='base', to='redmin.Domain', verbose_name='父模型')),
            ],
            options={
                'verbose_name': '字段',
                'verbose_name_plural': '字段',
                'ordering': ['base', 'group_sequence', 'sequence'],
            },
        ),
        migrations.CreateModel(
            name='Group',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50, verbose_name='名称')),
            ],
            options={
                'verbose_name': '用户组',
                'verbose_name_plural': '用户组',
                'ordering': ['-id'],
            },
        ),
        migrations.CreateModel(
            name='Menu',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True, verbose_name='名称')),
                ('sequence', models.IntegerField(default=0, verbose_name='排序')),
            ],
            options={
                'verbose_name': '菜单',
                'verbose_name_plural': '菜单',
                'ordering': ['sequence'],
            },
        ),
        migrations.CreateModel(
            name='SuperPermissionModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'verbose_name': '超級模型',
                'verbose_name_plural': '超級模型',
                'ordering': ['-id'],
            },
        ),
        migrations.CreateModel(
            name='UserFilter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, verbose_name='过滤器名称')),
                ('value_type', models.SmallIntegerField(choices=[(10, '固定值'), (5, '请求属性')], verbose_name='值类型')),
                ('attribute', models.CharField(max_length=100, verbose_name='属性名')),
                ('value', models.CharField(max_length=500, verbose_name='属性值')),
                ('enabled', models.BooleanField(default=True, verbose_name='开通')),
                ('domain', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Domain')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': '过滤器(用户)',
                'verbose_name_plural': '过滤器(用户)',
                'ordering': ['-id'],
            },
        ),
        migrations.CreateModel(
            name='GroupFilter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, verbose_name='过滤器名称')),
                ('value_type', models.SmallIntegerField(choices=[(10, '固定值'), (5, '请求属性')], verbose_name='值类型')),
                ('attribute', models.CharField(max_length=100, verbose_name='属性名')),
                ('value', models.CharField(max_length=500, verbose_name='属性值')),
                ('enabled', models.BooleanField(default=True, verbose_name='开通')),
                ('domain', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Domain')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Group')),
            ],
            options={
                'verbose_name': '过滤器(用户组)',
                'verbose_name_plural': '过滤器(用户组)',
                'ordering': ['-id'],
            },
        ),
        migrations.AddField(
            model_name='user',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Group'),
        ),
        migrations.AddField(
            model_name='user',
            name='groups',
            field=models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups'),
        ),
        migrations.AddField(
            model_name='user',
            name='user_permissions',
            field=models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions'),
        ),
        migrations.CreateModel(
            name='UserPermission',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creatable', models.BooleanField(default=False, verbose_name='可创建')),
                ('savable', models.BooleanField(default=False, verbose_name='可保存')),
                ('removable', models.BooleanField(default=False, verbose_name='可删除')),
                ('cloneable', models.BooleanField(default=False, verbose_name='可复制')),
                ('exportable', models.BooleanField(default=False, verbose_name='可导出')),
                ('viewable', models.BooleanField(default=False, verbose_name='可查看')),
                ('listable', models.BooleanField(default=False, verbose_name='可列表')),
                ('selectable', models.BooleanField(default=True, verbose_name='可选择')),
                ('domain', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Domain')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': '权限(用户)',
                'verbose_name_plural': '权限(用户)',
                'ordering': ['-id'],
                'unique_together': {('user', 'domain')},
            },
        ),
        migrations.CreateModel(
            name='UserField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=200, verbose_name='显示名称')),
                ('sequence', models.IntegerField(verbose_name='序号')),
                ('listable', models.BooleanField(default=True, verbose_name='在列表中显示')),
                ('formable', models.BooleanField(default=True, verbose_name='在表单中显示')),
                ('sortable', models.BooleanField(default=True, verbose_name='可排序')),
                ('exportable', models.BooleanField(default=True, verbose_name='可导出')),
                ('nullable', models.BooleanField(default=False, verbose_name='可以为空')),
                ('unique', models.BooleanField(default=False, verbose_name='是否唯一性')),
                ('default', models.CharField(blank=True, max_length=500, null=True, verbose_name='默认值')),
                ('form_editable', models.BooleanField(default=True, verbose_name='表单可编辑')),
                ('list_editable', models.BooleanField(default=False, verbose_name='列表可编辑')),
                ('searchable', models.BooleanField(default=False, help_text='通过一个关键字搜索选中的每一列', verbose_name='可搜索')),
                ('filterable', models.BooleanField(default=True, help_text='是否可在列表任务栏过滤', verbose_name='可过滤')),
                ('serializable', models.BooleanField(default=True, help_text='RestAPI里是否可以序列化', verbose_name='可序列化')),
                ('help_text', models.TextField(blank=True, null=True, verbose_name='帮助文本')),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Field')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': '字段(用户)',
                'verbose_name_plural': '字段(用户)',
                'ordering': ['user', 'field'],
                'unique_together': {('user', 'field')},
            },
        ),
        migrations.CreateModel(
            name='UserDomain',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=100, verbose_name='描述')),
                ('sequence', models.IntegerField(default=99999, verbose_name='排序')),
                ('app', models.CharField(max_length=100, verbose_name='应用')),
                ('name', models.CharField(max_length=100, verbose_name='名称')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': '用户模型',
                'verbose_name_plural': '用户模型',
                'ordering': ['user', 'sequence'],
                'unique_together': {('user', 'app', 'name')},
            },
        ),
        migrations.CreateModel(
            name='GroupPermission',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creatable', models.BooleanField(default=False, verbose_name='可创建')),
                ('savable', models.BooleanField(default=False, verbose_name='可保存')),
                ('removable', models.BooleanField(default=False, verbose_name='可删除')),
                ('cloneable', models.BooleanField(default=False, verbose_name='可复制')),
                ('exportable', models.BooleanField(default=False, verbose_name='可导出')),
                ('viewable', models.BooleanField(default=False, verbose_name='可查看')),
                ('listable', models.BooleanField(default=False, verbose_name='可列表')),
                ('selectable', models.BooleanField(default=True, verbose_name='可选择')),
                ('domain', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Domain')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Group')),
            ],
            options={
                'verbose_name': '权限(用户组)',
                'verbose_name_plural': '权限(用户组)',
                'ordering': ['-id'],
                'unique_together': {('group', 'domain')},
            },
        ),
        migrations.CreateModel(
            name='GroupField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=200, verbose_name='显示名称')),
                ('sequence', models.IntegerField(verbose_name='序号')),
                ('listable', models.BooleanField(default=True, verbose_name='在列表中显示')),
                ('formable', models.BooleanField(default=True, verbose_name='在表单中显示')),
                ('sortable', models.BooleanField(default=True, verbose_name='可排序')),
                ('exportable', models.BooleanField(default=True, verbose_name='可导出')),
                ('nullable', models.BooleanField(default=False, verbose_name='可以为空')),
                ('unique', models.BooleanField(default=False, verbose_name='是否唯一性')),
                ('default', models.CharField(blank=True, max_length=500, null=True, verbose_name='默认值')),
                ('form_editable', models.BooleanField(default=True, verbose_name='表单可编辑')),
                ('list_editable', models.BooleanField(default=False, verbose_name='列表可编辑')),
                ('searchable', models.BooleanField(default=False, help_text='通过一个关键字搜索选中的每一列', verbose_name='可搜索')),
                ('filterable', models.BooleanField(default=True, help_text='是否可在列表任务栏过滤', verbose_name='可过滤')),
                ('serializable', models.BooleanField(default=True, help_text='RestAPI里是否可以序列化', verbose_name='可序列化')),
                ('help_text', models.TextField(blank=True, null=True, verbose_name='帮助文本')),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Field', verbose_name='默认字段')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Group')),
            ],
            options={
                'verbose_name': '字段(用户组)',
                'verbose_name_plural': '字段(用户组)',
                'ordering': ['group', 'field'],
                'unique_together': {('group', 'field')},
            },
        ),
        migrations.CreateModel(
            name='GroupDomain',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=100, verbose_name='描述')),
                ('sequence', models.IntegerField(default=99999, verbose_name='排序')),
                ('app', models.CharField(max_length=100, verbose_name='应用')),
                ('name', models.CharField(max_length=100, verbose_name='名称')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Group')),
            ],
            options={
                'verbose_name': '用户组模型',
                'verbose_name_plural': '用户组模型',
                'ordering': ['group', 'sequence'],
                'unique_together': {('group', 'app', 'name')},
            },
        ),
        migrations.CreateModel(
            name='DomainMenu',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('domain', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Domain')),
                ('menu', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Menu')),
            ],
            options={
                'verbose_name': '模型菜单',
                'verbose_name_plural': '模型菜单',
                'ordering': ['menu__sequence', 'domain__sequence'],
                'unique_together': {('menu', 'domain')},
            },
        ),
        migrations.CreateModel(
            name='Choice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=100, verbose_name='标题')),
                ('value', models.CharField(max_length=100, verbose_name='值')),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='redmin.Field', verbose_name='默认字段')),
            ],
            options={
                'verbose_name': '选项',
                'verbose_name_plural': '选项',
                'ordering': ['field', 'id'],
                'unique_together': {('field', 'value')},
            },
        ),
    ]



from functools import wraps

from django.http import HttpResponse
from django.urls import path as django_patb
from django.views.decorators.csrf import csrf_exempt

from redmin.utils import *

methods = {}


def get_urlpatterns():
    return [
        django_patb(r, get_view(func), name=f"{func.__module__}.{func.__qualname__}")
        for r, func in methods.items()
    ]


def route(path, prefix=""):
    def decorate(func):
        new_path = prefix + path
        if new_path.startswith("/"):
            new_path = new_path[1:]

        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            return result

        methods[new_path] = wrapper
        return wrapper

    return decorate


def api_route(path, login_required=True):
    def decorate(func):
        new_path = path
        if new_path.startswith("/"):
            new_path = new_path[1:]

        @wraps(func)
        def wrapper(*args, **kwargs):
            if not login_required or args[0].user.is_authenticated:
                result = func(*args, **kwargs)
                if type(result) is not dict:
                    return HttpResponse(status=500, content="Result is not a dict obj")
                return json_response(func(*args, **kwargs))
            else:
                return HttpResponse(status=401)

        methods[new_path] = wrapper

    return decorate


def get_view(func):
    qualname = func.__qualname__
    module_name = func.__module__
    cls = None
    if "." in qualname:
        class_name, _ = qualname.split(".")
        cls = import_class("%s.%s" % (module_name, class_name))

        @csrf_exempt
        def view(request, *args, **kwargs):
            obj = cls()
            obj.request = request
            obj.args = args
            obj.kwargs = kwargs
            return func(obj, *args, **kwargs)
    else:
        def view(request, *args, **kwargs):
            return func(request, *args, **kwargs)

    return view



from django.urls import reverse
from rest_framework import permissions

from redmin.models import Permission, PermissionItem
from redmin.utils import attr


class RolePermission(permissions.IsAuthenticated):

    def has_permission(self, request, view):
        path = request.get_full_path()
        if path.startswith(reverse("account-login")) or path.startswith(reverse("account-logout")):
            return True
        model = attr(view, "model")
        if model:
            return Permission.has_permission(request.user, model, PermissionItem.viewable)
        return super().has_permission(request, view)



from django.db import models
from django.db.models.base import ModelBase
from django.urls import reverse

from redmin.utils import attr


def merge(name, attrs, meta_name, base_meta):
    meta = attrs.pop(meta_name, None)
    if not meta:
        meta = type(meta_name, (base_meta,), dict(
            __module__=f'{attrs.get("__module__")}.{name}'
        ))
    attributes = vars(meta)
    for key, value in vars(base_meta).items():
        if ("__init__" == key or not key.startswith("_")) and "abstract" != key and key not in attributes:
            setattr(meta, key, value)
    return meta


class BaseMeta(ModelBase):
    def __new__(mcs, name, bases, attrs, **kwargs):
        if name != 'RedminModel':
            attrs["Meta"] = merge(name, attrs, "Meta", getattr(RedminModel, "Meta"))
        return ModelBase.__new__(mcs, name, bases, attrs, **kwargs)


class RedminModel(models.Model, metaclass=BaseMeta):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    class Meta:
        ordering = ['-id']
        abstract = True

    def get_absolute_url(self):
        return reverse(self.__class__.__name__ + "-update", kwargs={"pk": self.pk})

    @property
    def css_color(self):
        return "black"

    @property
    def instance_permission(self):
        from .permission import Permission
        request = attr(self, "_request")
        if request:
            return Permission.get_permission(request.user, self.__class__).to_instance_permission()
        else:
            return Permission()

    @classmethod
    def get_domain_key(cls):
        app = cls.__module__.split(".")[0]
        name = cls.__name__
        return app + "_" + name



import logging
from functools import partial

import django.apps
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.middleware.csrf import get_token
from django.urls import path
from django.views.decorators.http import require_http_methods

from redmin.forms import LoginForm
from redmin.models import RedminModel
from redmin.utils import memorize, import_class
from .user import LoginView

logger = logging.getLogger(__name__)


@memorize
def get_base_view_class(app_name, action):
    try:
        return import_class(f"{app_name}.views.Base{action.capitalize()}View")
    except:
        return import_class(f"redmin.views.Admin{action.capitalize()}View")


def get_view(model, action):
    app_name = model.__module__.split(".")[0]

    view_name = "%s%sView" % (model.__name__, action.capitalize())
    view = None
    try:
        view = import_class(f'{app_name}.views.{view_name}')
    except:
        try:
            view = type("Dynamic%s" % view_name, (get_base_view_class(app_name, action),), dict(
                model=model,
                __module__=f'{app_name}.{__name__}',

            ))
        except Exception as e:
            logger.error(e)

    return view.as_view()


@login_required
@require_http_methods(["GET"])
def token_request(request):
    return HttpResponse(content=get_token(request))


@memorize
def get_models():
    return [model for model in django.apps.apps.get_models() if issubclass(model, RedminModel)]


def get_admin_domain_patterns(admin_prefix):
    urlpatterns = []
    for model in get_models():
        name = model.__name__
        prefix = f'{admin_prefix}/{name}'
        view = partial(get_view, model)
        urlpatterns += [
            path(f'{prefix}/', view("list"), name=f'{name}-list'),
            path(f'{prefix}/export/', view("export"), name=f'{name}-export'),
            path(f'{prefix}/create/', view("create"), name=f'{name}-create'),
            path(f'{prefix}/<int:pk>/', view("update"), name=f'{name}-update'),
            path(f'{prefix}/<int:pk>/delete/', view("delete"), name=f'{name}-delete'),
        ]

    urlpatterns.append(path(f'{admin_prefix}/token/', token_request))
    return urlpatterns


def get_admin_login_path(_route):
    return path(_route, LoginView.as_view(success_url='/'), kwargs={'authentication_form': LoginForm})



from django.db.models.signals import post_save
from django.urls import reverse
from django.views.generic.edit import UpdateView

from redmin.models import Permission, PermissionItem, Field
from redmin.utils import param, json_response, attr
from .form import AdminFormView


class AdminUpdateView(AdminFormView, UpdateView):
    success_message = "更新成功!"

    def get(self, request, *args, **kwargs):
        partial, attribute, value = param(request, ["partial", "attribute", "value"])
        pk = kwargs.get("pk")
        if partial and attribute and value and pk:
            instance = self.model.objects.filter(pk=pk).first()
            python_type = attr(Field.get_field(self.request.user, self.model, attribute), "python_type")
            if "django.db.models.fields.BooleanField" == python_type:
                value = True if value == "true" else False

            setattr(instance, attribute, value)
            # 用此方法只更新一个字段
            self.model.objects.bulk_update([instance], fields=[attribute])
            # bulk_update不发post_save信号
            post_save.send(self.model)
            return json_response({"success": True})

        return super().get(request, args, kwargs)

    def is_clone(self):
        return self.request.GET.get("clone") is not None

    def get_template_names(self):
        name = "create.html" if self.is_clone() else "update.html"
        return [f"admin/{self.model.__name__}/{name}", f'redmin/admin/base/{name}']

    def has_permission(self):
        return Permission.has_permission(self.request.user, self.model, PermissionItem.viewable)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['form_action'] = self.get_object().get_absolute_url()
        context["view_type"] = 'update'
        if self.is_clone():
            obj = context["object"]
            obj.id = ""
            context["form_action"] = reverse(f'{self.model.__name__}-create')
            context["is_clone"] = True

        setattr(context.get('object'), "_request", self.request)
        return context



import datetime


def attr(obj, attribute, default=None):
    name = str(attribute).strip()
    result = None
    if obj is not None:
        if "." not in name:
            try:
                if isinstance(obj, dict):
                    result = obj.get(name, default)
                elif isinstance(obj, (list, tuple)):
                    result = obj[int(name)]
                elif hasattr(obj, name):
                    result = getattr(obj, name)
            except Exception:
                result = None
        else:
            parts = name.split(".", 1)
            result = attr(attr(obj, parts[0]), parts[1])

    if result is None:
        result = default
    return result


def display(obj, attribute):
    from redmin.models import Field, Choice
    if obj:
        result = attr(obj, attribute, default="")
        field = Field.get_default_field(obj.__class__, attribute)
        if field:
            for choice in Choice.get_choices(field) or []:
                if choice.value == f"{result}":
                    return choice.title

        if isinstance(result, bool):
            result = "是" if result else '否'
        elif isinstance(result, datetime.datetime):
            result = result.strftime("%Y-%m-%d %H:%M:%S")
        return result
    return ""



from collections import OrderedDict

from django.db.models import Q
from django.forms import ChoiceField, forms
from django.views import generic
from django.views.generic.list import BaseListView

from redmin.models import Permission, PermissionItem, Filter, Field
from redmin.utils import param, json_response
from .mixins import ContextMixin, AccessMixin


class SearchMixin(BaseListView):
    def get_toolbar_search_params(self):
        fields = [field.attribute for field in Field.get_fields(self.request.user, self.model)]
        return {k: v for k, v in self.request.GET.items() if k in fields and v}

    def get_request_model_filters(self):
        fields = [field.attribute for field in Field.get_fields(self.request.user, self.model)]
        return {k: v for k, v in self.request.GET.items() if k in fields and v}


class FuzzySearchMixin(BaseListView):
    def get_queryset(self):
        queryset = super().get_queryset()
        keyword = self.request.GET.get("___fuzzy_search_key__")
        if keyword and len(keyword) > 0:
            search_attributes = Field.get_search_attributes(self.request.user, self.model)
            if search_attributes:
                q = Q()
                for attribute in search_attributes:
                    q = q | Q(**{f'{attribute.replace(".", "__")}__icontains': keyword})
                queryset = queryset.filter(q)

        return queryset


class FilterForm(forms.Form):
    pass


class ToolbarSearchFormMixin(SearchMixin):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["toolbar_search_form"] = self.get_toolbar_search_form()
        return context

    def get_toolbar_search_form(self):
        choices = self.get_toolbar_search_fields() + self.get_toolbar_extra_search_choices()
        if choices:
            fields = OrderedDict(choices)
            form = FilterForm()
            form.fields = fields
            return form

    def get_toolbar_extra_search_choices(self):
        return []

    def get_toolbar_search_fields(self):
        user = self.request.user
        choices = []
        group_fields = Field.get_group_fields(user, self.model, has_domain=True)
        params = self.get_toolbar_search_params()
        for fields in group_fields:
            fields = [field for field in fields if field.filterable]
            last_options = None
            last_default_value = None
            for index in range(len(fields)):
                field = fields[index]
                attribute = field.attribute
                cls = field.model
                queryset = None
                if index == 0:
                    queryset = cls.objects
                else:
                    last_attribute = fields[index - 1].attribute
                    last_value = params.get(last_attribute) or last_default_value
                    if last_value and last_options:
                        attribute_ = last_attribute[len(attribute) + 1:]
                        if attribute_:
                            filters = {f"{attribute_}_id": int(last_value)}
                            queryset = cls.objects.filter(**filters)

                queryset = Filter.filter(queryset, self.request, cls)
                options = [(e.id, str(e)) for e in queryset.all()] if queryset else []
                label = field.title
                if len(options) > 1:
                    options = [('', '选择%s' % label)] + options
                if options:
                    choices.append(
                        (attribute, ChoiceField(initial=params.get(attribute, ""), label=label, choices=options, )))

                last_default_value = options[0][0] if options else None
                last_options = options

        return choices


class OrderMixin(BaseListView):
    def get_queryset(self):
        queryset = super().get_queryset()
        sort = self.request.GET.get("sort")
        if sort:
            queryset = queryset.order_by(sort.replace('.', '__'))

        return queryset


class ToolbarSearchMixin(SearchMixin):
    def get_queryset(self):
        queryset = super().get_queryset()
        params = self.get_toolbar_search_params()
        fields_group = Field.get_group_fields(self.request.user, self.model, has_domain=True)
        for fields in fields_group:
            for field in fields:
                value = params.get(field.attribute, None)
                if value:
                    queryset = queryset.filter(**{field.attribute.replace(".", "__") + "_id": value})
                    break

        sort = self.request.GET.get("sort")
        if sort:
            queryset = queryset.order_by(sort.replace('.', '__'))

        return queryset


class ChoiceResponseMixin(BaseListView):
    def get(self, request, *args, **kwargs):
        choices, attribute, value = param(request, ["choices", "attribute", "value"])
        from redmin.models import Field
        field = Field.get_field(self.request.user, self.model, attribute)
        if choices and attribute and (value or field.nullable):
            queryset = self.model.objects
            if value:
                queryset = queryset.filter(**{attribute.replace(".", "__") + "_id": int(value)})
            queryset = Filter.filter(queryset, self.request, self.model)
            return json_response([{"id": obj.id, "title": str(obj)} for obj in queryset.all()], safe=False)
        return super().get(request, args, kwargs)


class AdminListView(
    ChoiceResponseMixin,
    FuzzySearchMixin,
    ToolbarSearchFormMixin,
    ToolbarSearchMixin,
    OrderMixin,
    ContextMixin,
    AccessMixin,
    generic.ListView
):
    context_object_name = 'list'
    paginate_by = 30
    form_class = None

    def get_model_list_fields(self):
        from redmin.models import Choice
        result = []
        for field in Field.get_fields(self.request.user, self.model, has_domain=False):
            if field.listable:
                field.choices = Choice.get_choices(field)
                result.append(field)
        return result

    def get_relation_fields(self):
        user = self.request.user
        fields = [field for field in Field.get_fields(user, self.model, has_domain=True) if
                  field.listable and Permission.has_permission(user, field.model, PermissionItem.listable)]
        return fields

    def get_list_fields(self):
        return self.get_relation_fields() + self.get_model_list_fields()

    def get_template_names(self):
        return [f'admin/{self.model.__name__}/list.html', 'redmin/admin/base/list.html']

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({"list_fields": self.get_list_fields()})
        context["view_type"] = 'list'
        for obj in context.get("list"):
            setattr(obj, "_request", self.request)

        return context

    def get_queryset(self):
        return Filter.filter(super().get_queryset(), self.request, self.model)

    def has_permission(self):
        item = PermissionItem.selectable if param(self.request, "choices") else PermissionItem.listable
        return Permission.has_permission(self.request.user, self.model, item)



import collections

import django.apps
from django.db import models
from filelock import FileLock

from redmin.utils import attr
from .base import RedminModel
from .field import Field

lock = FileLock("redmin_choice.lock")
lock.release(force=True)


def init_choices():
    with lock:
        news = []
        exists = collections.defaultdict(dict)
        for choice in Choice.objects.all():
            exists[choice.field][choice.value] = choice
        for model in django.apps.apps.get_models():
            if issubclass(model, RedminModel):
                for field in attr(model, '_meta.fields'):
                    choices = attr(field, "choices")
                    name = attr(field, "name")
                    if choices and name:
                        for value, title in choices:
                            f = Field.get_default_field(model, name)
                            if str(value) not in exists[f]:
                                news.append(Choice(field=f, value=str(value), title=title))
        if news:
            Choice.objects.bulk_create(news)


cache_lock = FileLock("redmin_choice_cache.lock")
cache_lock.release(force=True)

cache = collections.defaultdict(list)


def get_choices():
    if not cache:
        with cache_lock:
            for choice in Choice.objects.all():
                cache[choice.field].append(choice)

    return cache


class Choice(RedminModel):
    class Meta:
        ordering = ['field', "id"]
        verbose_name_plural = verbose_name = "选项"
        unique_together = [("field", "value")]

    field = models.ForeignKey(Field, on_delete=models.CASCADE, verbose_name="默认字段")
    title = models.CharField("标题", max_length=100)
    value = models.CharField("值", max_length=100)

    @classmethod
    def get_choices(cls, field):
        return get_choices()[field]

    def __str__(self):
        return f"{self.field.attribute},{self.value},{self.title}"



from collections import OrderedDict

from rest_framework.pagination import PageNumberPagination


class RedminPagination(PageNumberPagination):
    page_query_param = 'page'
    page_size_query_param = "pageSize"
    max_page_size = 50

    def get_paginated_response(self, data):
        from rest_framework.response import Response
        result = OrderedDict([
            ('pagination', {
                "total": self.page.paginator.count,
                "pageSize": self.page.paginator.per_page,
                "current": self.page.number
            }),
            ('list', data)
        ])
        return Response(result)

    def get_next_link(self):

        from rest_framework.utils.urls import replace_query_param
        if not self.page.has_next():
            return None
        url = self.request.build_absolute_uri()
        page_number = self.page.next_page_number()
        result = replace_query_param(url, self.page_query_param, page_number)
        result = replace_query_param(result, self.page_size_query_param, self.page.paginator.per_page)
        return result

    def get_previous_link(self):
        from rest_framework.utils.urls import replace_query_param, remove_query_param
        if not self.page.has_previous():
            return None
        url = self.request.build_absolute_uri()
        page_number = self.page.previous_page_number()

        if page_number == 1:
            result = remove_query_param(url, self.page_query_param)
        else:
            result = replace_query_param(url, self.page_query_param, page_number)

        result = replace_query_param(result, self.page_size_query_param, self.page.paginator.per_page)
        return result



from django.conf import settings
from django.contrib import auth
from django.contrib.auth import authenticate
from django.contrib.auth import logout as django_logout
from django.middleware.csrf import get_token
from rest_framework.decorators import api_view

from redmin.utils.rsa import decrypt as rsa_decrypt
from redmin.utils import json_response, attr


@api_view(['POST'])
def login(request):
    data = request.data
    username = data.pop("username", None)
    password = data.pop("password", None)
    encrypted = data.pop("encrypted", False)
    if encrypted:
        private_key = attr(settings, 'REMDIN_PRIVATE_KEY')
        if not private_key:
            return json_response({"code": 500, "status": "failed", "detail": "No private key found, please config REMDIN_PRIVATE_KEY"})
        password = rsa_decrypt(password, private_key)
    user = authenticate(request, username=username, password=password)
    if user:
        auth.login(request, user)
        return json_response({"type": "account", "currentAuthority": "admin", "code": 200, "status": "ok"})
    else:
        return json_response({"type": "account", "currentAuthority": "guest", "code": 403, "status": "failed"})


@api_view(['GET'])
def logout(request):
    django_logout(request)
    return json_response({"code": 200})


@api_view(['GET'])
def token(request):
    return json_response({"code": 200, "token": get_token(request=request), "status": "ok"})


@api_view(['GET'])
def profile(request):
    user = request.user
    return json_response({
        "code": 200,
        "status": "ok",
        "name": attr(user, "username", "guest"),
        "title": attr(user, "title", "访客"),
        "avatar": "https://gw.alipayobjects.com/zos/antfincdn/XAosXuNZyF/BiazfanxmamNRoxxVxka.png",
        "userid": attr(user, "id", -1),
        "email": attr(user, "email", "guest@guest.com"),
    })



import sys

from django.apps import AppConfig
from django.db.models.signals import post_migrate

from redmin.management import init_models


class RedminConfig(AppConfig):
    name = 'redmin'

    def ready(self):
        post_migrate.connect(init_models)




import django.dispatch

# box_changed = django.dispatch.Signal(providing_args=["before", "after"])
# order_created = django.dispatch.Signal(providing_args=["order"])
app_ready = django.dispatch.Signal(providing_args=[])



from django.urls import reverse
from django.views.generic.edit import CreateView

from redmin.models import Permission, PermissionItem
from .form import AdminFormView


class AdminCreateView(AdminFormView, CreateView):
    success_message = "创建成功!"

    def get_template_names(self):
        return [f"admin/{self.model.__name__}/create.html", 'redmin/admin/base/create.html']

    def has_permission(self):
        return Permission.has_permission(self.request.user, self.model, PermissionItem.creatable)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["form_action"] = reverse(f'{self.model.__name__}-create')
        context["view_type"] = 'create'

        return context



import datetime
import json
import logging
import traceback

from django.conf import settings
from django.http import HttpResponse
from django.utils.deprecation import MiddlewareMixin

from redmin.utils import attr, get_url_prefix

logger = logging.getLogger(__name__)

debug = attr(settings, 'DEBUG')


class LogMiddleware(MiddlewareMixin):
    def __init__(self, get_response=None):
        super().__init__(get_response)
        self.get_response = get_response

    def process_request(self, request):
        self.time = datetime.datetime.now()
        data = ""
        if request.method in ["POST", "PUT"]:
            post_data = request.POST
            if post_data:
                data = " form:%s" % json.dumps(request.POST)
            else:
                body = attr(request, "body")
                if body:
                    data = f" body:{body}"
        self.data = data
        pass

    def process_response(self, request, response):
        duration = (datetime.datetime.now() - self.time).microseconds
        prefix = get_url_prefix(request)
        content = ""
        content_type = response['Content-Type']
        if content_type.split(";")[0].lower() in ['application/json', "text/xml"] and response.content:
            content = "response:%s" % response.content.decode('raw_unicode_escape')

        name = attr(request, "user.username")
        user_info = f"user:{name}" if name else ""
        path = request.get_full_path()
        ip = request.META.get('HTTP_X_FORWARDED_FOR') or request.META.get('REMOTE_ADDR') or "no-ip"
        message = f'{ip} {request.method} "{prefix}{path}"{self.data} {response.status_code} {duration}ms {len(response.content)}b "{content_type}" {content} {user_info}'
        try:
            logger.info(message)
        except:
            pass
        return response

    def process_exception(self, request, exception):
        meta = request.META
        ip = meta.get('HTTP_X_FORWARDED_FOR') or meta.get('REMOTE_ADDR') or "no-ip"
        path = request.get_full_path()
        trace = traceback.format_exc()
        logger.error(f'{ip} {request.method} {path},process_exception: {json.dumps({"trace": trace})}')
        if debug:
            try:
                logger.error(trace)
            except Exception:
                pass

        if request.GET.get("debug") == '100':
            return HttpResponse(trace, content_type="application/json")



from collections import OrderedDict, defaultdict

from rest_framework.metadata import SimpleMetadata

from redmin.models import Field, Choice
from redmin.utils import attr


class RedminMetadata(SimpleMetadata):

    def determine_metadata(self, request, view):
        metadata = OrderedDict()
        metadata['name'] = view.get_view_name()
        metadata['description'] = view.get_view_description()
        metadata['renders'] = [renderer.media_type for renderer in view.renderer_classes]
        metadata['parses'] = [parser.media_type for parser in view.parser_classes]

        model = attr(view, "model")
        if model:
            fields = []
            choices = defaultdict(list)
            for choice in Choice.objects.filter(field__base__app=model.__module__.split(".")[0], field__base__name=model.__name__):
                choices[choice.field.attribute].append(OrderedDict(title=choice.title, value=choice.value))
            for field in Field.get_default_fields(view.model).values():
                if field.serializable and "." not in field.attribute:
                    attributes = OrderedDict(
                        name=field.attribute,
                        type=field.data_type,
                        label=field.title,
                        required=not field.nullable,
                        default=field.default,
                        sortable=field.sortable,
                        searchable=field.searchable,
                        description=field.help_text,
                        maxLength=field.max_length,
                        readOnly=not field.form_editable
                    )
                    field_choices = choices.get(field.attribute, None)
                    if field_choices:
                        attributes["choices"] = field_choices
                    fields.append(attributes)
            metadata['fields'] = fields
        return {"schema": metadata}



from rest_framework.renderers import BrowsableAPIRenderer as RestBrowsableAPIRenderer
from rest_framework.renderers import JSONRenderer as RestJSONRenderer

from redmin.utils import attr
from django.conf import settings


class BrowsableAPIRenderer(RestBrowsableAPIRenderer):
    template = 'redmin/rest/api.html'

    def get_context(self, data, accepted_media_type, renderer_context):
        context = super().get_context(data, accepted_media_type, renderer_context)
        context["redmin_page_title"] = attr(settings, "REST_HTML_TITLE", "Rest ")
        return context


class JSONRenderer(RestJSONRenderer):
    charset = 'utf-8'
    pass



inited = False


def init_models(sender, **kwargs):
    global inited
    if not inited:
        inited = True
        from redmin.models.field import init_fields
        from redmin.models.domain import init_domain
        from redmin.models.choice import init_choices
        type_map = init_domain()
        init_fields(type_map)
        init_choices()



from collections import defaultdict

import django.apps
from django.db import models
from django.db.models import ForeignKey
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from filelock import FileLock

from redmin.utils import first, attr
from .base import RedminModel
from .domain import Domain
from .group import Group
from .user import User

all_field_cache = defaultdict(lambda: defaultdict(list))
search_key_cache = defaultdict(lambda: defaultdict(list))
default_field_cache = defaultdict(dict)

all_field_lock = FileLock("all_field.lock")
all_field_lock.release(force=True)

search_locker = FileLock("redmin_field-search.lock")
search_locker.release(force=True)

default_field_lock = FileLock("default_field_lock.lock")
default_field_lock.release(force=True)


def get_all_search_attributes():
    if not search_key_cache:
        with search_locker:
            for model in django.apps.apps.get_models():
                if issubclass(model, RedminModel):
                    for user in User.objects.all():
                        for foreign_field in Field.get_fields(user, model, has_domain=True):
                            for field in Field.get_fields(user, foreign_field.model, has_domain=False):
                                if field.searchable:
                                    search_key_cache[user][model].append(f"{foreign_field.attribute}.{field.attribute}")

                        for field in Field.get_fields(user, model, has_domain=False):
                            if field.searchable:
                                search_key_cache[user][model].append(field.attribute)
    return search_key_cache


def get_default_fields():
    if not default_field_cache:
        with default_field_lock:
            for field in Field.objects.all():
                model = field.base.get_model()
                default_field_cache[model][field.attribute] = field
    return default_field_cache


def get_all_fields():
    if not all_field_cache:
        with all_field_lock:
            temp_cache = defaultdict(lambda: defaultdict(dict))
            for field in Field.objects.all():
                model = field.base.get_model()
                for user in User.objects.all():
                    temp_cache[user][model][field.attribute] = field

            base_attributes = ['sequence', 'listable', 'formable', 'sortable', 'exportable', 'title']
            for group_field in GroupField.objects.all():
                field = group_field.field
                model = field.base.get_model()
                for user in User.objects.filter(group=group_field.group).all():
                    default_field = temp_cache[user][model][field.attribute]
                    for attribute in base_attributes:
                        setattr(default_field, attribute, getattr(group_field, attribute))

            for user_field in UserField.objects.all():
                field = user_field.field
                model = field.base.get_model()
                default_field = temp_cache[user][model][field.attribute]
                for attribute in base_attributes:
                    setattr(default_field, attribute, getattr(user_field, attribute))

            for user, model_map in temp_cache.items():
                for model, field_dict in model_map.items():
                    all_field_cache[user][model] = sorted(field_dict.values(), key=lambda f: (f.group_sequence, f.sequence))

    return all_field_cache


class CommonField(RedminModel):
    class Meta:
        abstract = True

    title = models.CharField("显示名称", max_length=200)
    sequence = models.IntegerField("序号")
    listable = models.BooleanField("在列表中显示", default=True)
    formable = models.BooleanField("在表单中显示", default=True)
    sortable = models.BooleanField("可排序", default=True)
    exportable = models.BooleanField("可导出", default=True)
    nullable = models.BooleanField("可以为空", default=False)
    unique = models.BooleanField("是否唯一性", default=False)
    default = models.CharField("默认值", max_length=500, null=True, blank=True)
    form_editable = models.BooleanField("表单可编辑", default=True)
    list_editable = models.BooleanField("列表可编辑", default=False)
    searchable = models.BooleanField("可搜索", default=False, help_text="通过一个关键字搜索选中的每一列")
    filterable = models.BooleanField("可过滤", default=True, help_text="是否可在列表任务栏过滤")
    serializable = models.BooleanField("可序列化", default=True, help_text="RestAPI里是否可以序列化")
    help_text = models.TextField("帮助文本", null=True, blank=True)


DATA_TYPES = {
    'AutoField': 'integer',
    'BigAutoField': 'integer',
    'BooleanField': 'boolean',
    'CharField': 'string',
    'DateField': 'date',
    'DateTimeField': 'datetime',
    'DecimalField': 'float',
    'DurationField': 'integer',
    'FileField': 'string',
    'FilePathField': 'string',
    'FloatField': 'float',
    'IntegerField': 'integer',
    'BigIntegerField': 'integer',
    'IPAddressField': 'string',
    'GenericIPAddressField': 'string',
    'NullBooleanField': 'boolean',
    'OneToOneField': 'integer',
    'PositiveIntegerField': 'integer',
    'PositiveSmallIntegerField': 'integer',
    'SlugField': 'string',
    'SmallIntegerField': 'smallint',
    'TextField': 'text',
    'TimeField': 'time',
    'UUIDField': 'string',
}


class BaseField(RedminModel):
    """
    These Fields cannot be changed
    """

    class Meta:
        abstract = True

    base = models.ForeignKey(Domain, on_delete=models.CASCADE, verbose_name="父模型", related_name="base")
    attribute = models.CharField("字段名称", max_length=100)
    attribute_domain = models.ForeignKey(Domain, verbose_name="字段模型", null=True, blank=True, on_delete=models.CASCADE, related_name="attribute_domain")
    group_sequence = models.IntegerField("分组序号")
    python_type = models.CharField("原生类型", max_length=200)
    data_type = models.CharField("数据类型", max_length=10)
    max_length = models.IntegerField("最大长度", default=0)


class Field(BaseField, CommonField):
    class Meta:
        ordering = ['base', 'group_sequence', 'sequence']
        verbose_name_plural = verbose_name = "字段"

    def __str__(self):
        return f"{self.base},{self.title}({self.attribute})"

    @classmethod
    def get_default_fields(cls, model):
        return get_default_fields()[model]

    @classmethod
    def get_default_field(cls, model, attribute):
        return get_default_fields()[model].get(attribute)

    @classmethod
    def get_search_attributes(cls, user, model):
        return get_all_search_attributes()[user][model]

    @classmethod
    def get_field(cls, user, model, attribute):
        return first([field for field in get_all_fields()[user][model] if field.attribute == attribute])

    @classmethod
    def get_fields(cls, user, model, has_domain=None):
        result = []
        for field in get_all_fields()[user][model]:
            if has_domain is None or (has_domain is True and field.attribute_domain) or (
                    has_domain is False and not field.attribute_domain):
                result.append(field)
        return result

    @property
    def model(self):
        if self.attribute_domain:
            return self.attribute_domain.get_model()

    @property
    def class_name(self):
        return attr(self.model, "__name__")

    @classmethod
    def get_group_fields(cls, user, model, has_domain=None, reverse=False):
        result = []
        group_sequence = -1
        fields = []
        for field in cls.get_fields(user, model, has_domain):
            if group_sequence != field.group_sequence:
                group_sequence = field.group_sequence
                if fields:
                    result.append(fields)
                fields = []
            fields.append(field)
        if fields:
            result.append(fields)

        for i in range(len(result)):
            fields = sorted(result[i], key=lambda f: (f.group_sequence, f.sequence))
            result[i] = list(reversed(fields)) if reverse else fields

        return result


class GroupField(CommonField):
    class Meta:
        ordering = ["group", 'field']
        verbose_name_plural = verbose_name = "字段(用户组)"
        unique_together = [("group", "field")]

    group = models.ForeignKey(Group, on_delete=models.CASCADE)
    field = models.ForeignKey(Field, on_delete=models.CASCADE, verbose_name="默认字段")

    def __str__(self):
        return f"{self.group},{self.title}({self.field.attribute})"


class UserField(CommonField):
    class Meta:
        ordering = ["user", 'field']
        verbose_name_plural = verbose_name = "字段(用户)"
        unique_together = [("user", "field")]

    user = models.ForeignKey(User, on_delete=models.CASCADE)
    field = models.ForeignKey(Field, on_delete=models.CASCADE)

    def __str__(self):
        return f"{self.user},{self.title}({self.field.attribute})"


def get_attributes(cls, name=None):
    result = []
    foreign_fields = [(attr(f, "remote_field.model"), attr(f, "remote_field.field.name")) for f in
                      attr(cls, '_meta.fields') if issubclass(type(attr(f, "remote_field.field")), ForeignKey)]
    for foreign_model, foreign_attribute in foreign_fields:
        new_name = f"{name}.{foreign_attribute}" if name else foreign_attribute
        new_names = get_attributes(foreign_model, new_name)
        if new_names:
            result += new_names
        else:
            result.append(new_name)

    return result


def contains_chinese(word):
    for ch in word:
        if '\u4e00' <= ch <= '\u9fff':
            return True
    return False


def init_fields(type_map):
    basic_fields = [f.name for f in attr(BaseField, '_meta.fields')]
    new = []
    updates = []
    for model in django.apps.apps.get_models():
        if issubclass(model, RedminModel):
            group_sequence = 100
            base = type_map.get(model)
            attributes = []
            exists = {f.attribute: f for f in Field.objects.filter(base=base).all()}

            def check(field):
                exists_field = exists.get(field.attribute)
                if exists_field:
                    need_update = False
                    for key in basic_fields:
                        if attr(field, key) != attr(exists_field, key):
                            setattr(exists_field, key, attr(field, key))
                            need_update = True
                    if need_update:
                        updates.append(exists_field)
                        if len(updates) > 50:
                            Field.objects.bulk_update(updates, fields=basic_fields)
                            updates.clear()
                else:
                    new.append(field)
                    if len(new) > 50:
                        Field.objects.bulk_create(new)
                        new.clear()

            fields = [field for field in attr(model, '_meta.fields') if not attr(field, "remote_field")]
            for sequence, field in enumerate(fields, start=1):
                attribute = attr(field, "name")
                title = attr(field, '_verbose_name') or attribute
                attributes.append(attribute)
                field_type = type(field)
                python_type = f"{field_type.__module__}.{field_type.__name__}"
                form_editable = formable = attr(field, "form_editable", True) and attribute != "id"
                list_editable = attr(field, "list_editable", False) and attribute != "id"
                check(Field(
                    base=base,
                    attribute=attribute,
                    attribute_domain=None,
                    group_sequence=group_sequence,
                    python_type=python_type,
                    data_type=DATA_TYPES.get(python_type.split(".")[-1], "string"),
                    max_length=attr(field, "max_length") or 0,
                    serializable=True,
                    sequence=sequence * 5,
                    title=title,
                    nullable=attr(field, "null"),
                    form_editable=form_editable is not False,
                    list_editable=list_editable is not False,
                    formable=formable is not False,
                    help_text=attr(field, "help_text"),
                ))

            last_attribute, group_sequence = None, 0
            for attribute in get_attributes(model):
                names = attribute.split(".")
                for sequence in range(1, len(names) + 1):
                    sub_attribute, cls, title, field = ".".join(names[0:sequence]), model, None, None
                    if last_attribute and not sub_attribute.startswith(last_attribute):
                        group_sequence += 1
                    for name in sub_attribute.split("."):
                        field = attr(cls, f"{name}.field")
                        title = attr(field, "verbose_name")
                        cls = attr(field, f"remote_field.model")
                    form_editable = formable = attr(field, "form_editable", True) and sub_attribute != "id"
                    list_editable = attr(field, "list_editable", False) and sub_attribute != "id"
                    python_type = ForeignKey.__module__ + "." + ForeignKey.__name__
                    if not contains_chinese(title):
                        cls_verbose_name = attr(cls, "_meta.verbose_name")
                        if contains_chinese(cls_verbose_name):
                            title = cls_verbose_name
                    check(Field(
                        base=base,
                        group_sequence=group_sequence,
                        sequence=len(names) - sequence,
                        attribute=sub_attribute,
                        attribute_domain=type_map[cls],
                        title=title,
                        nullable=attr(field, "null"),
                        form_editable=form_editable is not False,
                        list_editable=list_editable is not False,
                        formable=formable is not False,
                        python_type=python_type,
                        data_type=DATA_TYPES.get(python_type.split(".")[-1], "string"),
                        max_length=attr(field, "max_length") or 0,
                        serializable=True,
                        help_text=attr(field, "help_text"),
                    ))
                    attributes.append(sub_attribute)
                    last_attribute = sub_attribute

            Field.objects.filter(base=base).exclude(attribute__in=attributes).delete()
    Field.objects.bulk_create(new)
    Field.objects.bulk_update(updates, fields=basic_fields)


@receiver([post_save, post_delete])
def handle_model_change(sender, **kwargs):
    if sender in [Group, User, Field, UserField, GroupField]:
        with all_field_lock:
            all_field_cache.clear()
            search_key_cache.clear()
            default_field_cache.clear()



from collections import OrderedDict
from urllib import parse

from django.urls import NoReverseMatch
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.routers import DefaultRouter, APIRootView

from redmin.models import Permission, PermissionItem
from redmin.rest.utils import get_all_model_map
from redmin.utils import get_url_prefix, attr

permission_items = [item for item in vars(PermissionItem).keys() if "_" not in item]


class _RootView(APIRootView):
    """
    Cinema Restful API
    """

    def get_view_name(self):
        return "影院后台列表"

    def get(self, request, *args, **kwargs):
        ret = OrderedDict()
        namespace = request.resolver_match.namespace
        url_prefix = get_url_prefix(request)
        all_models = get_all_model_map()
        for model_name, url_name in self.api_root_dict.items():
            model = all_models.get(model_name, None)
            permission: PermissionItem = Permission.get_permission(request.user, model)
            if not model or not permission or not permission.viewable:
                continue

            if namespace:
                url_name = namespace + ':' + url_name
            try:
                url = reverse(url_name, args=args, kwargs=kwargs, request=request, format=kwargs.get('format', None))
                parse_result = parse.urlparse(url)
                scheme, netloc = parse_result.scheme, parse_result.netloc
                ret[model_name] = {
                    "url": url.replace(f"{scheme}://{netloc}", url_prefix),
                    "permission": {item: attr(permission, item) for item in permission_items}
                }
            except NoReverseMatch:
                continue
        return Response(ret)


class BaseRouter(DefaultRouter):
    APIRootView = _RootView



import pkgutil


def import_class(name):
    import importlib
    module_name, class_name = name.rsplit(".", 1)
    module = importlib.import_module('.', module_name)
    return getattr(module, class_name)


def import_model(app_name, name):
    from django.apps import apps as django_apps
    if isinstance(name, str):
        result = None
        if "." in name:  # "hotel.Channel"
            result = django_apps.get_model(name)
        if not result:
            try:
                result = django_apps.get_model(app_name, name.lower())
            except Exception:
                pass
        if not result:
            try:
                result = import_class("%s.models.%s" % (app_name, name.capitalize()))
            except Exception:
                pass

        if result:
            return result
        raise Exception(f"Cannot import model by {app_name} {name}")


class ModuleProxyCache(dict):
    def __missing__(self, key):
        if '.' not in key:
            return __import__(key)

        module_name, class_name = key.rsplit('.', 1)

        module = __import__(module_name, {}, {}, [class_name])
        handler = getattr(module, class_name)

        # We cache a NoneType for missing imports to avoid repeated lookups
        self[key] = handler

        return handler


_cache = ModuleProxyCache()


def import_string(path):
    """
    Path must be module.path.ClassName

    >>> cls = import_string('sentry.models.Group')
    """
    result = _cache[path]
    return result


def import_submodules(context, root_module, path):
    """
    Import all submodules and register them in the ``context`` namespace.

    >>> import_submodules(locals(), __name__, __path__)
    """
    for loader, module_name, is_pkg in pkgutil.walk_packages(path, root_module + '.'):
        # this causes a Runtime error with model conflicts
        # module = loader.find_module(module_name).load_module(module_name)
        module = __import__(module_name, globals(), locals(), ['__name__'])
        for k, v in vars(module).items():
            if not k.startswith('_'):
                context[k] = v
        context[module_name] = module


def import_sub_classes(context, root_module, path):
    """
    Import all submodules and register them in the ``context`` namespace.

    >>> import_sub_classes(locals(), __name__, __path__)
    """
    for loader, module_name, is_pkg in pkgutil.walk_packages(path, root_module + '.'):
        module = __import__(module_name, globals(), locals(), ['__name__'])
        for k, v in vars(module).items():
            is_class = hasattr(v, '__name__')
            has_module = hasattr(v, "__module__")
            if not k.startswith('_') and is_class and has_module:
                name = v.__name__
                if name[0:1].isupper() and v.__module__.startswith(root_module):
                    context[k] = v
        context[module_name] = module



from redmin.utils import memorize


@memorize
def get_all_models():
    import django.apps
    from redmin.models import RedminModel
    return [model for model in django.apps.apps.get_models() if issubclass(model, RedminModel) and model.__module__.split(".")[0] != 'redmin']


@memorize
def get_all_model_map():
    return {model.__name__: model for model in get_all_models()}



from django.contrib.auth.forms import AuthenticationForm
from django.forms import widgets

app_name = __name__.split(".")[0]


class LoginForm(AuthenticationForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['username'].widget = widgets.TextInput(attrs={'placeholder': "username", "class": "form-control"})
        self.fields['password'].widget = widgets.PasswordInput(attrs={'placeholder': "password", "class": "form-control"})



from django.http import JsonResponse
from python_utils import converters


def param(request, key, default=None):
    if isinstance(key, list):
        return [param(request, item, default) for item in key]
    value = request.GET.get(key)
    if value is None:
        value = request.POST.get(key)
    if value is None:
        value = default
    return value


def int_param(request, param_name, default=0):
    return converters.to_int(param(request, param_name), default)


def bool_param(request, param_name):
    value = param(request, param_name, "")
    return value.upper() in ["TRUE", "1", "OK"]


def json_response(data, safe=True):
    return JsonResponse(data, safe=safe)


def get_url_prefix(request):
    meta = request.META
    host_name = meta.get('HTTP_X_HOST')
    scheme = meta.get('HTTP_X_SCHEME') or meta.get('wsgi.url_scheme')
    port = meta.get('HTTP_X_PORT')
    if host_name and port and scheme:
        port = '' if (scheme == 'https' and port == '443') or (scheme == 'http' and port == '80') else f':{port}'
        http_host = f'{host_name}{port}'
    else:
        http_host = meta.get('HTTP_HOST')
    prefix = f'{scheme}://{http_host}'
    return prefix



from django.http import HttpResponse
from django.template import loader


def render(request, template, context=None):
    return HttpResponse(loader.get_template(template).render(context or {}, request))



from django.db import models

from .base import RedminModel


class Group(RedminModel):
    class Meta:
        verbose_name_plural = verbose_name = "用户组"

    name = models.CharField('名称', max_length=50)

    def __str__(self):
        return self.name



import logging
from urllib import parse

from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from redmin.models import Field
from redmin.rest.serializer.base import load_serializer_class, get_dynamic_serializer_class
from redmin.utils import import_class, attr, get_url_prefix

logger = logging.getLogger(__name__)


class BaseViewSet(viewsets.ModelViewSet):
    model = None

    def get_extra_action_url_map(self):
        ret = super().get_extra_action_url_map()
        url_prefix = get_url_prefix(self.request)
        for key in ret.keys():
            url = ret[key]
            parse_result = parse.urlparse(url)
            scheme, netloc = parse_result.scheme, parse_result.netloc
            ret[key] = url.replace(f"{scheme}://{netloc}", url_prefix)
        return ret

    def get_serializer_class(self):

        fields = [field.attribute for field in Field.get_fields(self.request.user, self.model) if "." not in field.attribute]
        return get_dynamic_serializer_class(self.model, fields)

    @action(methods=['GET'], detail=False)
    def schema(self, request):
        meta = self.metadata_class()
        data = meta.determine_metadata(request, self)
        return Response(data)

    def get_queryset(self):
        queryset = super().get_queryset()
        return queryset

    def get_view_name(self):
        name = attr(self.model, "_meta.verbose_name")
        suffix = getattr(self, 'suffix', None)
        if suffix == "List":
            name += "(列表)"
        elif suffix == "Instance":
            name += "(实例)"

        return name


def load_viewset_class(model):
    app = model.__module__.split(".")[0]
    viewset_name = f"{model.__name__}ViewSet"
    serializer_class = load_serializer_class(model)
    search_fields = attr(model, "search_fields") or []
    filterset_fields = search_fields or serializer_class.Meta.fields
    try:
        clazz = import_class(f'{app}.rest.viewset.{viewset_name}')
        clazz.model = model
        if not attr(clazz, "serializer_class"):
            clazz.serializer_class = serializer_class
        if not attr(clazz, "queryset"):
            clazz.queryset = model.objects.all()
        if not attr(clazz, "filterset_fields"):
            clazz.filterset_fields = filterset_fields
        if not attr(clazz, "search_fields"):
            clazz.search_fields = search_fields
        logger.info(f"load rest viewset:{clazz}")
        return clazz
    except (AttributeError, ModuleNotFoundError):
        name = f"Dynamic{model.__name__}ViewSet"
        return type(name, (BaseViewSet,), dict(
            model=model,
            queryset=model.objects.all(),
            serializer_class=serializer_class,
            filterset_fields=filterset_fields,
            search_fields=search_fields,
            __module__=__name__
        ))



from django.core.exceptions import PermissionDenied
from django.http import Http404
from rest_framework import exceptions
from rest_framework.response import Response
from rest_framework.views import set_rollback


def exception_handler(exc, context):
    if isinstance(exc, Http404):
        exc = exceptions.NotFound()
    elif isinstance(exc, PermissionDenied):
        exc = exceptions.PermissionDenied()

    if isinstance(exc, exceptions.APIException):
        headers = {}
        if getattr(exc, 'auth_header', None):
            headers['WWW-Authenticate'] = exc.auth_header
        if getattr(exc, 'wait', None):
            headers['Retry-After'] = '%d' % exc.wait

        if isinstance(exc.detail, (list, dict)):
            response = Response(exc.detail, status=exc.status_code, headers=headers)
        else:
            response = Response({'detail': exc.detail, "code": exc.status_code, "status": "failed"}, headers=headers)

        set_rollback()
        return response

    return None



from collections import defaultdict

from django.db import models
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from filelock import FileLock

from redmin.utils import attr
from .base import RedminModel
from .domain import Domain
from .field import Field
from .group import Group
from .user import User

lock = FileLock("filter.lock")
lock.release(force=True)

cache = defaultdict(lambda: defaultdict(list))


@receiver([post_save, post_delete])
def clear_filter(sender, **kwargs):
    if sender in [Group, User] or issubclass(sender, Filter):
        with lock:
            cache.clear()


def get_all_filters():
    if not cache:
        with lock:
            for f in GroupFilter.objects.all():
                for user in User.objects.filter(group=f.group).all():
                    cache[user][f.domain.get_model()].append(f)

            for f in UserFilter.objects.all():
                cache[f.user][f.domain.get_model()].append(f)

    return cache


class FilterValueType:
    constant = 10
    request_attribute = 5

    choices = [
        (constant, "固定值"),
        (request_attribute, "请求属性"),
    ]


class Filter(RedminModel):
    class Meta:
        abstract = True

    domain = models.ForeignKey(Domain, on_delete=models.CASCADE)
    name = models.CharField("过滤器名称", max_length=100)
    value_type = models.SmallIntegerField("值类型", choices=FilterValueType.choices)
    attribute = models.CharField("属性名", max_length=100)
    value = models.CharField("属性值", max_length=500)
    enabled = models.BooleanField("开通", default=True)

    @classmethod
    def get_filters_dict(cls, request, user, model):
        result = {}
        for f in get_all_filters()[user][model]:
            value = cls.parse_value(request, f)
            if value is not None:
                result[f.attribute] = value
        return result

    @classmethod
    def parse_value(cls, request, f):
        return attr(request, f.value) if f.value_type == FilterValueType.request_attribute else f.value

    @classmethod
    def filter(cls, query, request, model):
        if query:
            user = request.user
            filters = cls.get_filters_dict(request, user, model)
            for fields in Field.get_group_fields(user, model, has_domain=True, reverse=True):
                for field in fields:
                    for f in get_all_filters()[user][field.model]:
                        value = cls.parse_value(request, f)
                        if value is not None:
                            filters[f"{field.attribute.replace('.', '__')}_{f.attribute}"] = value
            if filters:
                query = query.filter(**filters)
        return query


class GroupFilter(Filter):
    class Meta:
        verbose_name_plural = verbose_name = "过滤器(用户组)"

    group = models.ForeignKey(Group, on_delete=models.CASCADE)


class UserFilter(Filter):
    class Meta:
        verbose_name_plural = verbose_name = "过滤器(用户)"

    user = models.ForeignKey(User, on_delete=models.CASCADE)



import random
import string


def is_empty(s):
    return s is None or s.strip() == ""


def is_not_empty(s):
    return not is_empty(s)


def random_string(length=10):
    return ''.join(random.sample(string.ascii_letters + string.digits, length))



from .admin import AdminCreateView, AdminDeleteView, AdminExportView, AdminFormView, AdminListView, AdminUpdateView
from .admin import get_view as get_admin_view
from .admin.user import UserCreateView, UserUpdateView



import django.apps
from django.db import models
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from filelock import FileLock

from redmin.utils import attr
from .base import RedminModel
from .group import Group
from .user import User

lock = FileLock("redmin_domain.lock")
lock.release(force=True)

cache = dict()


@receiver([post_save, post_delete])
def handle_model_change(sender, **kwargs):
    if sender in [Domain, GroupDomain, UserDomain]:
        with lock:
            cache.clear()


def init_domain():
    from redmin.models import Domain, RedminModel
    domains = {}
    for domain in Domain.objects.all():
        domains[domain.app + "-" + domain.name] = domain
    all_models = {}
    for model in [model for model in django.apps.apps.get_models() if issubclass(model, RedminModel)]:
        app = model.__module__.split(".")[0]
        name = model.__name__
        all_models[app + "-" + name] = model

    type_map = {}
    for flag, model in all_models.items():
        domain = domains.get(flag)
        title = model._meta.verbose_name
        if domain:
            if domain.title != title:
                domain.title = title
                domain.save()
        else:
            app, name = flag.split("-")
            domain = Domain.objects.create(app=app, name=name, title=title)
        type_map[model] = domain
    for flag, domain in domains.items():
        if flag not in all_models:
            domain.delete()

    return type_map


def get_map():
    from collections import defaultdict
    if not cache:
        with lock:
            domains = {domain.get_key(): domain for domain in Domain.objects.all()}
            cache['domains'] = domains
            cache['models'] = {model.get_domain_key(): model for model in django.apps.apps.get_models() if
                               issubclass(model, RedminModel)}

            group_domains = defaultdict(dict)
            for c in GroupDomain.objects.all():
                group_domains[str(c.group.id)][c.get_key()] = c
            cache['group_domains'] = group_domains

            user_domains = defaultdict(dict)
            for c in UserDomain.objects.all():
                user_domains[str(c.user.id)][c.get_key()] = c
            cache['user_domains'] = group_domains

    return cache


@receiver(post_save)
@receiver(post_delete)
def clear_map(sender, **kwargs):
    if sender in [Group, User] or issubclass(sender, BaseDomain):
        with lock:
            cache.clear()


class BaseDomain(RedminModel):
    class Meta:
        abstract = True

    title = models.CharField("描述", max_length=100)
    sequence = models.IntegerField("排序", default=99999)
    app = models.CharField("应用", max_length=100)
    name = models.CharField("名称", max_length=100)

    def __str__(self):
        return self.title

    @classmethod
    def get(cls, app, name):
        return cls.get_by_key(f"{app}.{name}")

    @classmethod
    def get_by_user_and_model(cls, user, model):
        cache = get_map()
        key = model.get_domain_key()
        return attr(cache, f'user_domains.{user.id}.{key}') or attr(cache, f'group_domains.{user.group_id}.{key}') or attr(cache, f'domains.{key}')

    @classmethod
    def get_by_model(cls, model):
        return get_map()['domains'][model.get_domain_key()]

    def get_model(self):
        return get_map()['models'][self.get_key()]

    def get_key(self):
        return self.app + "_" + self.name

    @classmethod
    def get_by_key(cls, key):
        return get_map()['domains'][key]

    @classmethod
    def get_model_by_key(cls, key):
        return get_map()['models'][key]


class Domain(BaseDomain):
    class Meta:
        ordering = ['sequence', "id"]
        verbose_name_plural = verbose_name = "模型"
        unique_together = [("app", "name")]


class GroupDomain(BaseDomain):
    class Meta:
        ordering = ['group', 'sequence']
        verbose_name_plural = verbose_name = "用户组模型"
        unique_together = ("group", "app", "name")

    group = models.ForeignKey(Group, on_delete=models.CASCADE)


class UserDomain(BaseDomain):
    class Meta:
        ordering = ['user', 'sequence']
        verbose_name_plural = verbose_name = "用户模型"
        unique_together = ("user", "app", "name")

    user = models.ForeignKey(User, on_delete=models.CASCADE)



import logging

from django.core.management.base import BaseCommand

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def handle(self, *args, **options):
        self.init_fields()
        self.init_menu()
        self.create_user("dev", "dev", "123456")
        self.create_user("root", "root", "123456", True)

    @staticmethod
    def init_menu():
        from redmin.models import Menu, Domain, DomainMenu
        menu = Menu.objects.update_or_create(name="ROOT")[0]
        domains = list(Domain.objects.filter(app='redmin').all())
        sequence = 0
        for domain in domains:
            DomainMenu.objects.update_or_create(domain=domain, menu=menu)
            domain.sequence = sequence
            sequence += 5
        Domain.objects.bulk_update(domains, fields=['sequence'])

    @staticmethod
    def init_fields():
        from redmin.models import Field
        fields = list(Field.objects.filter(base__name="User"))
        for field in fields:
            flag = field.attribute in ['group', 'usename', 'title']
            field.formable = field.listable = flag
            field.list_editable = field.form_editable = False
        Field.objects.bulk_update(fields, fields=['formable', 'listable', 'list_editable', 'form_editable'])

        fields = list(Field.objects.filter(base__name="Field"))
        for field in fields:
            flag = field.attribute.endswith("able") or field.attribute.endswith("sequence")
            field.formable = field.list_editable = flag
            field.listable = field.attribute not in ['python_type', 'help_text', 'default', 'max_length', 'nullable', 'exportable', 'data_type', 'unique']
            if field.attribute == "group_sequence":
                field.sequence = 4

        Field.objects.bulk_update(fields, fields=['formable', 'listable', 'list_editable', 'sequence'])

        fields = list(Field.objects.filter(attribute="id"))
        for field in fields:
            field.listable = False
        Field.objects.bulk_update(fields, fields=['listable'])

    @staticmethod
    def create_user(group_name, username, password, super_user=False):
        from redmin.models import Group, User, GroupPermission, SuperPermissionModel, Domain
        root_group = Group.objects.update_or_create(name=group_name)[0]
        user = User.objects.filter(username=username).first()
        if not user:
            user = User(group=root_group, username=username, title=username)
            user.set_password(password)
            user.save()
            logger.info(f"root user created:username={username},password={password}")

        if super_user:
            domain = Domain.objects.filter(name=SuperPermissionModel.__name__).first()
            permission = GroupPermission.objects.filter(group=root_group, domain=domain).first()
            if not permission:
                GroupPermission.objects.create(
                    group=root_group,
                    domain=domain,
                    creatable=True,
                    savable=True,
                    removable=True,
                    cloneable=True,
                    exportable=True,
                    viewable=True,
                    listable=True,
                )



from django import forms
from django.conf import settings
from django.contrib import auth
from django.contrib.auth import REDIRECT_FIELD_NAME, logout
from django.contrib.auth.forms import AuthenticationForm
from django.utils.decorators import method_decorator
from django.utils.http import is_safe_url
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.debug import sensitive_post_parameters
from django.views.generic import FormView

from redmin.forms import LoginForm
from redmin.models import User
from redmin.utils import attr
from .create import AdminCreateView
from .update import AdminUpdateView

app_name = __name__.split(".")[0]


class Mixin(object):
    model = User

    def form_valid(self, form):
        user = form.save(False)
        password = form.cleaned_data['password1']
        if password and len(password) > 0:
            user.set_password(password)
            user.save()
        return super().form_valid(form)

    def get_form_class(self):
        form_class = super().get_form_class()

        class UserForm(form_class):
            password1 = forms.CharField(label='修改密码', min_length=6, max_length=30, widget=forms.PasswordInput, required=False, help_text=u"如果需要更改密码,请直接填写. 否则就留空")
            password2 = forms.CharField(label='确认密码', min_length=6, max_length=30, widget=forms.PasswordInput, required=False)

            def clean_password2(self):
                password1 = self.cleaned_data.get('password1')
                password2 = self.cleaned_data.get('password2')
                if password1 and len(password1) > 0 and password1 != password2:
                    raise forms.ValidationError("两次密码输入不一致!")
                return password2

        return UserForm


class UserCreateView(Mixin, AdminCreateView):
    pass


class UserUpdateView(Mixin, AdminUpdateView):
    pass


class LoginView(FormView):
    form_class = LoginForm
    template_name = 'redmin/admin/user/login.html'
    success_url = '/'
    redirect_field_name = REDIRECT_FIELD_NAME

    def get(self, request, *args, **kwargs):
        self.request = request
        logout(request)
        return super().get(request, *args, **kwargs)

    @method_decorator(sensitive_post_parameters('password'))
    @method_decorator(csrf_protect)
    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        redirect_to = self.request.GET.get(self.redirect_field_name)
        if redirect_to is None:
            redirect_to = attr(settings, 'INDEX_URL')
        kwargs['redirect_to'] = redirect_to

        return super().get_context_data(**kwargs)

    def form_valid(self, form):
        form = AuthenticationForm(data=self.request.POST, request=self.request)

        if form.is_valid():
            # redirect_to = self.request.GET.get(self.redirect_field_name)
            auth.login(self.request, form.get_user())
            return super().form_valid(form)
            # return HttpResponseRedirect('/')
        else:
            return self.render_to_response({'form': form})

    def get_success_url(self):
        redirect_to = self.request.POST.get(self.redirect_field_name)
        if not is_safe_url(url=redirect_to, allowed_hosts=[self.request.get_host()]):
            redirect_to = self.success_url
        return redirect_to
